#include <cstddef>
#include <cstdlib>
#include <iostream>

#include <CL/sycl.hpp>

constexpr auto dim_x = std::size_t{512};
constexpr auto dim_y = std::size_t{256};

constexpr auto local_dim_x = std::size_t{dim_x + 2}; // 0 left and right
constexpr auto local_dim_y = std::size_t{3};

// constexpr auto img_num = 512; // 2GiB input / output
constexpr auto img_num = std::size_t{3};

struct box_filter_t;
using box_filter = cl::sycl::xilinx::reqd_work_group_size<1, 1, 1, box_filter_t>;

struct xilinx_selector : public cl::sycl::device_selector
{
    auto operator()(const cl::sycl::device& dev) const -> int
    {
        const auto vendor = dev.get_info<cl::sycl::info::device::vendor>();
        return (vendor.find("Xilinx") != std::string::npos) ? 1 : -1;
    }
};

auto main() -> int
{
    try
    {
        auto exception_handler = [] (cl::sycl::exception_list exceptions)
        {
            for(std::exception_ptr e : exceptions)
            {
                try
                {
                    std::rethrow_exception(e);
                }
                catch(const cl::sycl::exception& err)
                {
                    std::cerr << "Caught asynchronous SYCL exception: "
                              << err.what() << std::endl;
                    std::exit(EXIT_FAILURE);
                }
            }
        };

        auto queue = cl::sycl::queue{xilinx_selector{}, exception_handler,
                                     cl::sycl::property::queue::enable_profiling{}};


        const auto buf_range = cl::sycl::range<2>{dim_x, dim_y};

        auto inputs = std::vector<cl::sycl::buffer<int, 2>>{};
        auto outputs = std::vector<cl::sycl::buffer<int, 2>>{};
        for(auto i = 0; i < img_num; ++i)
        {
            inputs.emplace_back(cl::sycl::buffer<int, 2>{buf_range});
            queue.submit([&](cl::sycl::handler& cgh)
            {
                auto acc = inputs.back().get_access<cl::sycl::access::mode::discard_write>(cgh);
                cgh.fill(acc, 42);
            });

            outputs.emplace_back(cl::sycl::buffer<int, 2>{buf_range});
            queue.submit([&](cl::sycl::handler& cgh)
            {
                auto acc = outputs.back().get_access<cl::sycl::access::mode::discard_write>(cgh);
                cgh.fill(acc, 0);
            });
        }

        for(auto i = 0; i < img_num; ++i)
        {
            queue.submit([&](cl::sycl::handler& cgh)
            {
                auto buffers_in = inputs.at(i).get_access<cl::sycl::access::mode::read>(cgh);
                auto buffers_out = outputs.at(i).get_access<cl::sycl::access::mode::write>(cgh);

                cgh.single_task<box_filter>(
                [=]()
                {
                    auto upper_row = cl::sycl::xilinx::partition_array<
                        int, local_dim_x,
                        cl::sycl::xilinx::partition::cyclic<local_dim_x / 2, 1>>{};
                    auto center_row = cl::sycl::xilinx::partition_array<
                        int, local_dim_x,
                        cl::sycl::xilinx::partition::cyclic<local_dim_x / 2, 1>>{};
                    auto lower_row = cl::sycl::xilinx::partition_array<
                        int, local_dim_x,
                        cl::sycl::xilinx::partition::cyclic<local_dim_x / 2, 1>>{};
                    auto out_row = cl::sycl::xilinx::partition_array<
                        int, dim_x,
                        cl::sycl::xilinx::partition::cyclic<dim_x / 2, 1>>{};

                    // for all rows
                    for(auto y = 0ul; y < dim_y; ++y)
                    {
                        // init local memory
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto xs = 0ul; xs < local_dim_x; ++xs)
                            {
                                upper_row[xs] = 0.f;
                            }
                        });
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto xs = 0ul; xs < local_dim_x; ++xs)
                            {
                                center_row[xs] = 0.f;
                            }
                        });
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto xs = 0ul; xs < local_dim_x; ++xs)
                            {
                                lower_row[xs] = 0.f;
                            }
                        });

                        // burst read three rows 
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto x = 0ul, xs = 1ul;
                                x < dim_x;
                                ++x, ++xs)
                            {
                                if(y == 0ul)
                                    break;

                                const auto global = cl::sycl::id<2>{x, y - 1ul};
                                upper_row[xs] = buffers_in[global];
                            }
                        });
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto x = 0ul, xs = 1ul;
                                x < dim_x;
                                ++x, ++xs)
                            {
                                const auto global = cl::sycl::id<2>{x, y};
                                center_row[xs] = buffers_in[global];
                            }
                        });
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto x = 0ul, xs = 1ul;
                                x < dim_x;
                                ++x, ++xs)
                            {
                                if(y == (dim_y - 1ul))
                                    break;

                                const auto global = cl::sycl::id<2>{x, y + 1ul};
                                lower_row[xs] = buffers_in[global];
                            }
                        });

                        // calculate box filter
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto x = 0ul, xs = 1ul;
                                x < dim_x;
                                ++x, ++xs)
                            {
                                const auto upper = upper_row[xs - 1ul] + upper_row[xs] + upper_row[xs + 1ul];
                                const auto center = center_row[xs - 1ul] +  center_row[xs] + center_row[xs + 1ul];
                                const auto lower = lower_row[xs - 1ul] + center_row[xs] + center_row[xs + 1ul];

                                out_row[x] = (upper + center + lower) / 9.f;
                            }
                        });

                        // burst write result
                        cl::sycl::xilinx::pipeline([&]()
                        {
                            for(auto x = 0ul; x < dim_x; ++x)
                            {
                                const auto global = cl::sycl::id<2>{x, y};
                                buffers_out[global] = out_row[x];
                            }
                        });
                   }
                });
            });
        }
        queue.wait_and_throw();

        auto host_out = outputs.back().get_access<cl::sycl::access::mode::read>();
        for(auto y = std::size_t{0}; y < std::size_t{3}; ++y)
        {
            for(auto x = std::size_t{0}; x < std::size_t{32}; ++x)
            {
                const auto id = cl::sycl::id<2>{x, y};
                std::cout << host_out[id] << " ";
            }
            std::cout << std::endl; 
        }
    }
    catch(const cl::sycl::exception& err)
    {
        std::cerr << "Caught synchronous SYCL exception: " << err.what() << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
